--[[-------------------------------------------------------------------------
-- Simple Interrupt Announce by Saphod - Draenor EU
-- All rights reserved
-------------------------------------------------------------------------]]--

local addonName, addon = ...
addon.version = GetAddOnMetadata(addonName, 'Version')

-- Expose our addon table through a global variable
_G[addonName] = addon

-- Create frame for handling events
addon.frame = CreateFrame('Frame', addonName .. 'Frame')
addon.frame:Hide()

-- Configuration table with default values
-- Note: Valid values for own and other are: off, self, say, party, raid
addon.cfg = {
	cfgversion = 1,
	isEnabled = true,
	-- When solo, how to announce your own interrupts and those of others
	solo = { own = 'self', other = 'off' },
	-- When in a party, how to announce your own interrupts and those of others
	party = { own = 'self', other = 'off' },
	-- When in a raid, how to announce your own interrupts and those of others
	raid = { own = 'say', other = 'self' },
}

-- Function to enable addon
function addon:Enable()
	if not self.cfg.isEnabled then
		self.cfg.isEnabled = true
		self.frame:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end
end

-- Function to disable addon
function addon:Disable()
	if self.cfg.isEnabled then
		self.cfg.isEnabled = false
		self.frame:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end
end

-- Local shorthands for accessing the frame and configuration
local frame = addon.frame
local cfg = addon.cfg

-- Local names for globals used in CLEU handler
local GetSpellLink = GetSpellLink
local UnitInParty = UnitInParty
local UnitInRaid = UnitInRaid
local UnitName = UnitName
local UnitPlayerOrPetInParty = UnitPlayerOrPetInParty
local UnitPlayerOrPetInRaid = UnitPlayerOrPetInRaid

-- Update configuration table
-- Note: Assumes all keys are simple and values are tables or simple.
-- Will overwrite a non-table with a table, but not the other way around.
local function UpdateCfg(dst, src)
	for k, v in pairs(src) do
		if type(v) == 'table' then
			if type(dst[k]) == 'table' then
				-- Update a table with a table
				UpdateCfg(dst[k], v)
			else
				-- Overwrite a value with a table
				dst[k] = UpdateCfg({}, v)
			end
		elseif type(dst[k]) ~= 'table' then
			-- Update a value with a value
			dst[k] = v
		end
	end
	return dst
end

-- Split a string into whitespace delimited parts and return them as a list
-- Note: Simple and slow, but we only use it to parse the slash commands.
local function SplitString(s)
	local t = {}
	-- Loop over matches of non-whitespace sequences in s, adding them to t
	for w in s:gmatch('%S+') do
		t[#t + 1] = w
	end
	return unpack(t)
end

-- Get name of player
-- Note: We cannot read the name into a local variable on load since
-- UnitName may return UNKNOWNOBJECT until the player has fully loaded.
-- We could potentially use the GUID instead.
local GetPlayerName
do
	local playerName

	function GetPlayerName()
		if not playerName then
			local a = UnitName('player')
			if a and a ~= UNKNOWNOBJECT then
				playerName = a
			end
		end
		return playerName
	end
end

-- Get spell link for a spellID
local GetSpellLinkCached
do
	local spellLinkCache = {}

	function GetSpellLinkCached(spellID)
		local a = spellLinkCache[spellID]
		if not a then
			a = GetSpellLink(spellID)
			spellLinkCache[spellID] = a
		end
		return a
	end
end

-- Event dispatcher
frame:SetScript('OnEvent', function(self, event, ...)
	local a = self[event]
	if a then
		a(self, ...)
	end
end)

-- Perform actions once all saved variables are loaded
function frame:ADDON_LOADED(name)
	if name == addonName then
		-- We do not have to be called for any further addons
		self:UnregisterEvent('ADDON_LOADED')

		-- Update current configuration table with saved
		if not SIACFG then SIACFG = {} end
		UpdateCfg(cfg, SIACFG)

		-- Register combat log event handling if announcing is on
		if cfg.isEnabled then
			self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
		end
	end
end

-- Perform actions as player is about to enter world
function frame:PLAYER_LOGIN()
	-- Add slash command
	SLASH_SIA1 = '/sia'
	SlashCmdList['SIA'] = function(msg)
		self:OnSlashCmd(SplitString((msg or ''):lower()))
	end

	-- Notify player we're loaded and how to get options
	self:Print(addonName, addon.version, 'loaded -- /sia for options')
end

-- Perform actions as player is about to leave world
function frame:PLAYER_LOGOUT()
	-- Update saved configuration table with current
	UpdateCfg(SIACFG, cfg)
end

-- The actual combat log event handler
-- Note: Naming the args was faster than select, and Lua adjusts the number
-- of arguments automatically.
function frame:COMBAT_LOG_EVENT_UNFILTERED(_, subEvent, _, _, sourceName, _, _, _, _, _, _, _, _, _, extraSpellID)
	-- Check if event was a spell interrupt
	if subEvent ~= 'SPELL_INTERRUPT' then return end

	-- Figure out grouping status
	local inParty, inRaid = UnitInParty('player'), UnitInRaid('player')

	local config

	-- Select config based on grouping status
	if inRaid then
		config = cfg.raid
	elseif inParty then
		config = cfg.party
	else
		config = cfg.solo
	end

	-- Check if source was the player
	if sourceName == GetPlayerName() then
		-- If own is off for this config, return
		if config.own == 'off' then return end

		-- Announce interrupt
		self:AnnounceInterrupt(config.own, 'Interrupted ' .. GetSpellLinkCached(extraSpellID))
	else
		-- If other is off for this config, return
		if config.other == 'off' then return end

		local isGrouped = false

		-- Check if source was our pet or a player/pet in our party/raid
		-- Note: UnitPlayerOrPetInX return nil for the player and his pet,
		-- so we check for our pet explicitly.
		if inRaid then
			if UnitPlayerOrPetInRaid(sourceName) or sourceName == UnitName('pet') then isGrouped = true end
		elseif inParty then
			if UnitPlayerOrPetInParty(sourceName) or sourceName == UnitName('pet') then isGrouped = true end
		else
			if sourceName == UnitName('pet') then isGrouped = true end
		end

		-- Announce if source was in our party/raid or our own pet
		if isGrouped then
			self:AnnounceInterrupt(config.other, (sourceName or '?') .. ' interrupted ' .. GetSpellLinkCached(extraSpellID))
		end
	end
end

-- Register the events we handle
frame:RegisterEvent('ADDON_LOADED')
frame:RegisterEvent('PLAYER_LOGIN')
frame:RegisterEvent('PLAYER_LOGOUT')

-- Print with SIA prefix
function frame:Print(...)
	return print('|cFFF0F050SIA|r:', ...)
end

-- Formatted print with SIA prefix
function frame:Printf(s, ...)
	return print('|cFFF0F050SIA|r:', string.format(s, ...))
end

-- Announce message to self or channel based on mode
-- Note: mode has to be self, say, party, or raid
function frame:AnnounceInterrupt(mode, msg)
	if mode == 'self' then
		self:Print(msg)
	else
		-- Here mode is say, party, or raid
		SendChatMessage(msg, mode:upper())
	end
end

-- Print a line with own and other modes in config
function frame:PrintMode(mode, config)
	self:Print('|cFF20E020' .. mode .. '|r modes: own -> |cFF20E020' .. config.own .. '|r, other -> |cFF20E020' .. config.other .. '|r')
end

-- Update config with the modes from own and other
function frame:ParseCmdModes(cmd, config, own, other)
	-- Update own mode
	if own then
		if own == 'off' then
			config.own = 'off'
		elseif own == 'self' then
			config.own = 'self'
		elseif own == 'say' then
			config.own = 'say'
		elseif own == 'party' and (cmd == 'party' or cmd == 'raid') then
			config.own = 'party'
		elseif own == 'raid' and cmd == 'raid' then
			config.own = 'raid'
		else
			self:Print('invalid mode for |cFF20E020' .. cmd .. '|r own:', own)
			return
		end
	end

	-- Update other mode
	if other then
		if other == 'off' then
			config.other = 'off'
		elseif other == 'self' then
			config.other = 'self'
		elseif other == 'say' then
			config.other = 'say'
		elseif other == 'party' and (cmd == 'party' or cmd == 'raid') then
			config.other = 'party'
		elseif other == 'raid' and cmd == 'raid' then
			config.other = 'raid'
		else
			self:Print('invalid mode for |cFF20E020' .. cmd .. '|r other:', other)
			return
		end
	end
end

-- Slash command handler
function frame:OnSlashCmd(cmd, ...)
	if cmd == 'on' then
		-- Register combat log event handling to start announcing
		self:Print('Announce turned |cFF20E020on|r (mode |cFF20E020' .. cfg.mode .. '|r)')
		addon:Enable()
	elseif cmd == 'off' then
		-- Unregister combat log event handling to stop announcing
		self:Print('Announce turned |cFFE02020off|r')
		addon:Disable()
	elseif cmd == 'solo' then
		self:ParseCmdModes(cmd, cfg.solo, ...)
		self:PrintMode(cmd, cfg.solo)
	elseif cmd == 'party' then
		self:ParseCmdModes(cmd, cfg.party, ...)
		self:PrintMode(cmd, cfg.party)
	elseif cmd == 'raid' then
		self:ParseCmdModes(cmd, cfg.raid, ...)
		self:PrintMode(cmd, cfg.raid)
	elseif cmd == 'info' then
		-- Show a little profiling info
		UpdateAddOnMemoryUsage()
		local usedKB = GetAddOnMemoryUsage(addonName)
		self:Printf('Mem usage %.2f KB', usedKB)
		if GetCVar('scriptProfile') == '1' then
			UpdateAddOnCPUUsage()
			local addOnTime = GetAddOnCPUUsage(addonName)
			self:Printf('AddOn CPU usage %.2f (%.2f%%)', addOnTime, (100 * addOnTime) / GetScriptCPUUsage())
			local frameTime, frameCount = GetFrameCPUUsage(frame)
			self:Printf('Frame CPU usage %.2f in %d calls (%.4f per call)', frameTime, frameCount, frameTime / frameCount)
		end
	elseif cmd == 'help' then
		-- Show an explanation of status and modes
		self:Print('For each status (|cFF20E020solo party|r and |cFF20E020raid|r), own and other')
		self:Print('mode can be set to |cFF20E020off self say party|r or |cFF20E020raid|r.')
		self:Print('Other announces interrupts of your pet or players/pets')
		self:Print('in your party/raid.')
		self:Print('Examples:')
		self:Print('  |cFF20E020/sia party|r - print current party modes')
		self:Print('  |cFF20E020/sia solo self|r - when solo, own interrupts to yourself')
		self:Print('  |cFF20E020/sia raid say self|r - when raid, own to say, others to self')
	else
		-- By default, show status and options
		self:Print(addonName, addon.version)
		self:Print('Syntax: /sia <command> [options]')
		self:Print('  |cFF20E020on/off|r - turn announcing on/off')
		self:Print('  |cFF20E020solo/party/raid [own [other]]|r - set mode')
		self:Print('  |cFF20E020help|r - show more help')
		if cfg.isEnabled then
			self:Print('Announcing is |cFF20E020on|r')
		else
			self:Print('Announcing is |cFFE02020off|r')
		end
	end
end
